import py3mf_slicer
import py3mf_slicer.load
import py3mf_slicer.slice
import py3mf_slicer.get_items
import py3mf_slicer.visualize
geometry1 = r"tests\geometries\test_geometry1.stl"
geometry2 = r"tests\geometries\test_geometry2.stl"
geometry3 = r"tests\geometries\test_geometry3.stl"

model = py3mf_slicer.load.load_files([geometry1, geometry2, geometry3])
        
sliced_model = py3mf_slicer.slice.slice_model(model, 1)

import time
start_time = time.time()  # Record start time
slices1 = py3mf_slicer.get_items.get_shapely_slice(sliced_model, 0)
end_time = time.time()    # Record end time

execution_time = end_time - start_time
print(f"Execution time1: {execution_time:.4f} seconds")

start_time = time.time()  # Record start time
slices2 = py3mf_slicer.get_items.get_shapely_slice2(sliced_model, 0)
end_time = time.time()    # Record end time

execution_time = end_time - start_time
print(f"Execution time2: {execution_time:.4f} seconds")

print(slices2)


import lib3mf
import pyvista as pv
from py3mf_slicer.get_items import get_pyvista_meshes
from ctypes import c_float, c_uint32
import numpy as np
import networkx as nx
import math

def to_lib3mf_position2D(positions):
    lib3mf_positions = []
    for row in positions:
        position = lib3mf.Position2D((c_float * 2)(row[0], row[1]))
        lib3mf_positions.append(position)
    return lib3mf_positions

def identify_pv_polygons(arr):
    from collections import defaultdict, deque
    # Create a graph from the array
    graph = defaultdict(list)
    for edge in arr:
        graph[edge[0]].append(edge[1])
        graph[edge[1]].append(edge[0])
    def bfs(start, visited):
        queue = deque([start])
        circle = []
        while queue:
            node = queue.popleft()
            if node not in visited:
                visited.add(node)
                circle.append(node)
                for neighbor in graph[node]:
                    if neighbor not in visited:
                        queue.append(neighbor)
        circle.append(start)  # Add the start node at the end to complete the circle
        return circle
    visited = set()
    circles = []
    for node in graph:
        if node not in visited:
            circle = bfs(node, visited)
            if len(circle) > 2:  # Ensure it's a circle
                circles.append(circle)
    return circles

def identify_pv_polygons2(connections):
    # Create a graph from the connections
    G = nx.Graph()
    G.add_edges_from(connections)
    
    # Find all connected components in the graph
    connected_components = list(nx.connected_components(G))
    
    sorted_entities = []
    
    for component in connected_components:
        # Create a subgraph for each component
        subgraph = G.subgraph(component)
        
        # Find a path that visits all nodes in the subgraph
        path = list(nx.dfs_preorder_nodes(subgraph))
        
        # Append the sorted path to the result
        sorted_entities.append(path)
    
    return sorted_entities
    
def slice_pv_mesh(mesh, layer_height):
    z_min, z_max = mesh.bounds[4], mesh.bounds[5]
    # Create a MultiBlock to store all slices
    slices = pv.MultiBlock()
    # Slice the mesh at each z-coordinate
    z_floor = math.floor(z_min/layer_height)*layer_height
    z_floor = max(z_floor, 0)
    z = z_floor+layer_height
    while z < z_max:
        slice = mesh.slice(normal=[0, 0, 1], origin=[0, 0, z])
        slices.append(slice)
        z = z + layer_height

    return slices

def slice_model(model, layer_height):
    model_sliced = model #copy.deepcopy(model)
    pv_meshes = get_pyvista_meshes(model_sliced)
    for mesh in pv_meshes:
        slices = slice_pv_mesh(mesh, layer_height)

        z_min = slices[0].points[0][2]
        z_max = slices[-1].points[0][2]

        slicestack = model_sliced.AddSliceStack(z_min-layer_height)
        for pv_slice in slices:
            slice = slicestack.AddSlice(pv_slice.points[0][2])
            vertices = pv_slice.points[:, :2]
            positions = to_lib3mf_position2D(vertices)
            slice.SetVertices(positions)
            pv_lines = pv_slice.lines.reshape(-1, 3)[:, 1:]
            #sorted_lines = identify_pv_polygons(pv_lines)
            sorted_lines = identify_pv_polygons2(pv_lines)

            for sorted_line in sorted_lines:
                sorted_line = np.array(sorted_line)
                lines = (c_uint32 * sorted_line.size)(*sorted_line.flatten())
                slice.AddPolygon(lines)
    return model_sliced




import py3mf_slicer.get_items as get_items

def transform_model(model, item, movement = (0, 0, 0), rotation=(0, 0, 0), scale=(1, 1, 1)):
    meshes = get_items.get_pyvista_meshes(model)
    mesh = meshes[item]
    transformed_mesh = mesh.translate([movement[0], movement[1], movement[2]])
    transformed_mesh = transformed_mesh.rotate_x(rotation[0]).rotate_y(rotation[1]).rotate_z(rotation[2])
    transformed_mesh = transformed_mesh.scale([scale[0], scale[1], scale[2]])

    meshes[item] = transformed_mesh
    import pyvista as pv
    plotter = pv.Plotter()
    for slice in meshes:
        plotter.add_mesh(slice, color='green')
    plotter.show()
    transformed_model = get_items.get_py3mf_from_pyvista(meshes)

    return transformed_model



import unittest
import py3mf_slicer
import py3mf_slicer.load
import py3mf_slicer.slice
import py3mf_slicer.get_items
import py3mf_slicer.visualize

class TestPy3MF(unittest.TestCase):

    def setUp(self):
        # Paths to the files used in the test
        self.geometry1 = r"tests\geometries\test_geometry1.stl"
        self.geometry2 = r"tests\geometries\test_geometry2.stl"
        self.geometry3 = r"tests\geometries\test_geometry3.stl"

    def test_load_and_slice_model(self):
        # Step 1: Load the model
        model = py3mf_slicer.load.load_files([self.geometry1, self.geometry2, self.geometry3])
        
        # Assert that the model is not None
        self.assertIsNotNone(model, "Model failed to load")
        
        # Step 2: Slice the model
        sliced_model = py3mf_slicer.slice.slice_model(model, 1)
        
        # Assert that slicing produced a valid object
        self.assertIsNotNone(sliced_model, "Slicing the model failed")
        
        # Step 3: Get slices
        slices = py3mf_slicer.get_items.get_slices(sliced_model)

        # Assert that slices were extracted
        self.assertGreater(len(slices), 0, "No slices were extracted from the model")

        # Step 4: Get shapely slices
        slices = py3mf_slicer.get_items.get_shapely_slice(sliced_model, 0)

        # Assert that slices were extracted
        self.assertGreater(len(slices), 0, "No shapely slices were extracted from the model")
        
        # Step 5: Test if correct number of slice stacks and mesh objects are returned
        number_of_mesh_objects = py3mf_slicer.get_items.get_number_of_mesh_objects(sliced_model)
        number_of_slice_stacks = py3mf_slicer.get_items.get_number_of_slice_stacks(sliced_model)
        number_of_layers = py3mf_slicer.get_items.get_number_layers(sliced_model)
        self.assertEqual(number_of_mesh_objects, 3, "Not correct number of mesh objects")
        self.assertEqual(number_of_slice_stacks, 3, "Not correct number of slice stacks")
        self.assertEqual(len(number_of_layers), 3, "Not correct number of slices")

        # Vizulise slice
        #py3mf_slicer.visualize.visualize_layer(slices[5])

        # Test get shapely slice
        exported_counts = [0] * len(number_of_layers)
        for i in range(max(number_of_layers)):
            slice = py3mf_slicer.get_items.get_shapely_slice(sliced_model, i)
            for i, value in enumerate(slice):
                if value is not None:
                    exported_counts[i] += 1  # Increment count for the column if value is not None
        self.assertEqual(exported_counts, number_of_layers, "Not correct number of exported shapely slices")

        
if __name__ == '__main__':
    unittest.main()


import lib3mf
import numpy as np
import pyvista as pv
from ctypes import c_float, c_uint32
import shapely

def to_lib3mf_position2D(positions):
    lib3mf_positions = []
    for row in positions:
        position = lib3mf.Position2D((c_float * 2)(row[0], row[1]))
        lib3mf_positions.append(position)
    return lib3mf_positions

def get_vertexs(model):
    matrix = []
    mesh_iterator = model.GetMeshObjects()
    
    while mesh_iterator.MoveNext():
        vertex_matrix = []
        mesh_object = mesh_iterator.GetCurrentMeshObject()
        vertices = mesh_object.GetVertices()
        
        for vertex in vertices:
            coordinates = vertex.Coordinates
            vertex_matrix.append([coordinates[0], coordinates[1], coordinates[2]])
        matrix.append(vertex_matrix)
    return matrix

def get_bounding_boxes(model):
    matrices = get_vertexs(model)
    boundries = []
    for matrix in matrices:
        m = np.array(matrix)
        min_values = np.min(m, axis=0)
        max_values = np.max(m, axis=0)
        boundries.append([max_values.tolist(), min_values.tolist()])
    return boundries

def get_layer_height(model):
    slice_stack_iterator = model.GetSliceStacks()
    while slice_stack_iterator.MoveNext():
        slicestack = slice_stack_iterator.GetCurrentSliceStack()
        slice_count = slicestack.GetSliceCount()
        z_min = slicestack.GetSlice(0).GetZTop()
        z_max = slicestack.GetSlice(slice_count-1).GetZTop()
        if slice_count > 1:
            layer_height = (z_max-z_min)/(slice_count-1)
        else:
            layer_height = z_min
        return layer_height
    return 0
            
def get_pyvista_meshes(model):
    pv_meshes = []
    mesh_iterator = model.GetMeshObjects()
    while mesh_iterator.MoveNext():
        mesh_object = mesh_iterator.GetCurrentMeshObject()
        # Extract vertices
        vertices = mesh_object.GetVertices()
        points = [(vertex.Coordinates[0], vertex.Coordinates[1], vertex.Coordinates[2]) for vertex in vertices]
        # Extract faces
        triangles = mesh_object.GetTriangleIndices()
        faces = []
        for triangle in triangles:
            faces.append([3, triangle.Indices[0], triangle.Indices[1], triangle.Indices[2]])
        # Create pyvista mesh
        pv_meshes.append(pv.PolyData(points, faces))
    return pv_meshes

def get_pyvista_slices(model):
    slice_stack_iterator = model.GetSliceStacks()
    slices = []
    while slice_stack_iterator.MoveNext():
        slicestack = slice_stack_iterator.GetCurrentSliceStack()
        multiblock = pv.MultiBlock()

        for i in range(slicestack.GetSliceCount()):
            slice = slicestack.GetSlice(i)
            vertices = []
            vertices_list = slice.GetVertices()
            z_top = slicestack.GetSlice(i).GetZTop()
            vertices = [[vertex.Coordinates[0], vertex.Coordinates[1], z_top] for vertex in vertices_list]
            
            polygons = []
            for k in range(slice.GetPolygonCount()):
                polygon = slice.GetPolygonIndices(k)
                newlst = []
                for i in range(len(polygon)-1):
                    newlst.append(2)
                    newlst.append(polygon[i])
                    newlst.append(polygon[i+1])
                newlst.append(2)
                newlst.append(polygon[-1])
                newlst.append(polygon[0])
                polydata = pv.PolyData(vertices, lines=np.array(newlst))
                multiblock.append(polydata)
            
        slices.append(multiblock)
    return slices

def get_py3mf_from_pyvista(pyvista_meshes):
    wrapper = lib3mf.get_wrapper()
    model = wrapper.CreateModel()
    for mesh in pyvista_meshes:
        mesh_object = model.AddMeshObject()
        vertices = mesh.points
        faces = mesh.faces
        # Add vertices to the lib3mf mesh object
        for vertex in vertices:
            mesh_object.AddVertex(lib3mf.Position((c_float * 3)(vertex[0], vertex[1], vertex[2])))
        # Add triangles to the lib3mf mesh object
        for i in range(0,len(faces),4):
            mesh_object.AddTriangle(lib3mf.Triangle((c_uint32 * 3)(faces[i+1], faces[i+2], faces[i+3])))
        transform = lib3mf.Transform()
        transform.m_Fields = [
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
            [0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0]
        ]
        model.AddBuildItem(mesh_object, transform)
    return model

def get_slices(model):
    pv_model_slices = get_pyvista_slices(model)
    levels = []
    max_z = 0
    for pv_model in pv_model_slices:
        level =[]
        for slice in pv_model:
            level.append(slice.GetBounds()[4])
        np_level = np.array(level)
        levels.append(np_level)
        max_z = max(max_z, np.max(np_level))

    layer_height = get_layer_height(model)
    z = layer_height
    slices = []

    while z <= max_z:
        slice = []
        for i in range(len(pv_model_slices)):
            positions = np.where(levels[i] == z)[0]
            s = []
            for pos in positions:
                block = pv_model_slices[i][pos.item()] #[pos]
                s.append(block)
            slice.append(s)
        slices.append(slice)
        z += layer_height
    return slices
      
def get_shapely_slice(model, layer):
    def create_hierarchy(polygons):
        polygons.sort(key=lambda p: p.area, reverse=True)
        top_level = [polygons[0]]
        second_level = [[]]
        for i in range(1,len(polygons)):
            added = False
            for ii in range(len(top_level)):
                if top_level[ii].contains(polygons[i]):
                    add_to_second_level = True
                    for iii in range(len(second_level[ii])):
                        if second_level[ii][iii].contains(polygons[i]):
                            add_to_second_level = False
                    if add_to_second_level:
                        second_level[ii].append(polygons[i])
                        added = True
            if not added:
                top_level.append(polygons[i])
                second_level.append([])
        polygons = []
        for i in range(len(top_level)):
            if len(second_level[i])>0:
                interiors = [polygon.exterior for polygon in second_level[i]]
                polygons.append(shapely.Polygon(top_level[i].exterior, interiors))
            else:
                polygons.append(top_level[i])
        return shapely.MultiPolygon(polygons)
    # Get bounding boxes and z bounds for each slicestack
    bounding_box = get_bounding_boxes(model)
    z_bound = np.array([[sublist[0][2], sublist[1][2]] for sublist in bounding_box])
    # Calculate layer height
    layer_height = get_layer_height(model)
    height = layer * layer_height
    # Initialize slice stack iterator
    slice_stack_iterator = model.GetSliceStacks()
    slicestacks = []
    # Safely traverse the slice stack iterator
    while slice_stack_iterator.MoveNext():
        try:
            slicestack = slice_stack_iterator.GetCurrentSliceStack()
            slicestacks.append(slicestack)
        except Exception as e:
            print(f"Error accessing slice stack: {e}")
            continue  # Skip to the next slice stack if there's an issue
    shapely_slice = []
    # Iterate over all slicestacks
    for index, slicestack in enumerate(slicestacks):
        # Calculate the relative height for this slicestack
        if z_bound[index][1] <= height < z_bound[index][0]-layer_height:
            # Only process slicestacks within bounds
            stack_layer = int(height - z_bound[index][1])
            slice = slicestack.GetSlice(stack_layer)
            vertices_list = slice.GetVertices()
            # Convert vertex coordinates to a NumPy array (faster than list comprehensions)
            np_vertices = np.array([[vertex.Coordinates[0], vertex.Coordinates[1]] for vertex in vertices_list])
            # Use list comprehension to create polygons efficiently
            polygons = [
                shapely.Polygon(np_vertices[slice.GetPolygonIndices(k)]) for k in range(slice.GetPolygonCount())
            ]
            # Add the polygons to the result for this slicestack
            if len(polygons) > 1:
                shapely_slice.append(create_hierarchy(polygons))
            else:
                shapely_slice.append(polygons[0])
        else:
            # If the slicestack is out of bounds, append an empty list
            shapely_slice.append(None)
    return shapely_slice

def get_number_of_mesh_objects(model):
    mesh_iterator = model.GetMeshObjects()
    nmb_of_meshes = 0
    while mesh_iterator.MoveNext():
        nmb_of_meshes = nmb_of_meshes + 1
    return nmb_of_meshes

def get_number_of_slice_stacks(model):
    slice_stack_iterator = model.GetSliceStacks()
    nmb_of_slice_stacks = 0
    while slice_stack_iterator.MoveNext():
        nmb_of_slice_stacks = nmb_of_slice_stacks + 1
    return nmb_of_slice_stacks

def get_number_layers(model):
    number_of_slices = []
    slice_stack_iterator = model.GetSliceStacks()
    while slice_stack_iterator.MoveNext():
        slicestack = slice_stack_iterator.GetCurrentSliceStack()
        slice_count = slicestack.GetSliceCount()
        number_of_slices.append(slice_count)
    return number_of_slices


from lib3mf import get_wrapper
import lib3mf
import os

def load_file(path):
    file_extension = os.path.splitext(path)[1].lower()
    if file_extension[1:] == 'stl' or file_extension[1:] == 'obj'or file_extension[1:] == '3mf':
        wrapper = get_wrapper()
        model = wrapper.CreateModel()
        reader = model.QueryReader(file_extension[1:])
        reader.ReadFromFile(path)
        return model
    else:
        print("Import file format not supported")
    return None

def load_files(paths):
    def add_mesh_objects(source_model, target_model):
        mesh_iterator = source_model.GetMeshObjects()
        while mesh_iterator.MoveNext():
            mesh_object = mesh_iterator.GetCurrentMeshObject()
            new_mesh_object = target_model.AddMeshObject()
            
            # Copy vertices and triangles
            vertices = mesh_object.GetVertices()
            triangles = mesh_object.GetTriangleIndices()
            new_mesh_object.SetGeometry(vertices, triangles)
            
            # Add the new mesh object to the build
            transform = lib3mf.Transform()
            transform.m_Fields = [
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0]
            ]
            target_model.AddBuildItem(new_mesh_object, transform)
            
    wrapper = lib3mf.get_wrapper()
    merged_model = wrapper.CreateModel()
    for path in paths:
        add_mesh_objects(load_file(path), merged_model)
    return merged_model
    



from setuptools import setup, find_packages

setup(
    name='py3mf_slicer',
    version='0.0.2',
    packages=find_packages(),
    install_requires=[
        'setuptools',
        'lib3mf',
        'pyvista',
        'numpy',
        'networkx',
        'shapely'
    ],
    entry_points={
        'console_scripts': [
            # Define command-line scripts here
        ],
    },
    author='Anton Wiberg',
    author_email='wiberg.anton@gmail.com',
    description='Functions which aids the process of using and slicing 3mf files',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/wiberganton/py3mf_slicer',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import pyvista as pv

def get_colors():
    colors = [
        'red', 'blue', 'green', 'yellow', 'orange', 'purple', 'pink', 'brown',
        'black', 'white', 'gray', 'cyan', 'magenta', 'lightblue', 'lightgreen',
        'lightgray', 'darkblue', 'darkgreen', 'darkgray', 'darkred', 'darkorange',
        'darkviolet', 'darkcyan', 'darkmagenta', 'skyblue', 'turquoise', 'lime',
        'olive', 'teal', 'navy', 'maroon', 'gold', 'khaki', 'orchid', 'plum',
        'salmon', 'sienna', 'indigo', 'chocolate', 'coral', 'crimson', 'lavender',
        'thistle', 'beige', 'hotpink', 'tan', 'forestgreen', 'seagreen',
        'royalblue', 'midnightblue', 'slateblue', 'deepskyblue', 'dodgerblue',
        'lightsalmon', 'lightcoral', 'peachpuff', 'palegreen', 'palevioletred',
        'springgreen', 'steelblue', 'tomato', 'wheat', 'azure', 'bisque',
        'blanchedalmond', 'cornflowerblue', 'gainsboro', 'honeydew', 'ivory',
        'lavenderblush', 'lemonchiffon', 'linen', 'mistyrose', 'oldlace',
        'papayawhip', 'peachpuff', 'seashell', 'snow', 'whitesmoke', 'aliceblue'
    ]
    return colors

## Excpects input on the form [[pv_PolyData, pv_PolyData], [pv_PolyData], [...]]
## will assign one color to each inner list
def visualize_layer(pv_layer):
    plotter = pv.Plotter()
    colors = get_colors()
    for i, polydata in enumerate(pv_layer):
        for polydatum in polydata:
            plotter.add_mesh(polydatum, color=colors[i], show_edges=True)
    plotter.show()